(require 'macro)
(require 'sort)
(require 'printf)

;;; dates
(define (make-date yyyy mm dd)
  (let ((t (make-vector 3)))
    (begin
      (vector-set! t 0 yyyy)
      (vector-set! t 1 mm)
      (vector-set! t 2 dd)
      t )))

(define (date->yyyy t) (vector-ref t 0))
(define (date->mm t) (vector-ref t 1))
(define (date->dd t) (vector-ref t 2))

(define (date-less? a b)
  (or (< (date->yyyy a) (date->yyyy b))
      (and (= (date->yyyy a) (date->yyyy b))
	   (or (< (date->mm a) (date->mm b))
	       (and (= (date->mm a) (date->mm b))
		    (< (date->dd a) (date->dd b)))))))

;; list functions from Olin's library
(define (filter pred lis)
  (let recur ((lis lis))
    (if (pair? lis)
	(let ((head (car lis))
	      (tail (cdr lis)))
	  (if (pred head)
	      (let ((new-tail (recur tail)))
		(if (eq? tail new-tail) lis
		    (cons head new-tail)))
	      (recur tail)))
	'())))

(define (filter* preds lis)
  (if (null? preds)
      lis
      (filter* (cdr preds) (filter (car preds) lis))))

;;; values
;; Values are lists of pairs (currency . amount), as in
;; '((USD . 10) (Eur . 20)) .  Algebraically, values are
;; first-degree polynomials in the currencies.

; merge terms with the same currency
(define (value/canonicalize val)
  (define (sum l x) (if (null? l) x (sum (cdr l) (+ x (car l)))))

  (if (null? val)
      val
      (let* ((currency (caar val))
	     (same (filter (lambda (v) (eq? currency (car v))) val))
	     (other (filter (lambda (v) (not (eq? currency (car v)))) val)))
	(cons (cons currency (sum (map cdr same) 0))
	      (value/canonicalize other)))))

; add list of values
(define (value/sum l)
  (value/canonicalize (apply append l)))

(define (make-currency c)
  (let ((in-currency (cons 'in c))
	(out-currency (cons 'out c)))
    (lambda (amount)
      (lambda (return)
	(return (cons in-currency amount) (cons out-currency amount))))))
      
;;; transactions
(define (make-transaction amount date description other-account reconciled?)
  (let ((t (make-vector 5)))
    (begin
      (vector-set! t 0 amount)
      (vector-set! t 1 date)
      (vector-set! t 2 description)
      (vector-set! t 3 other-account)
      (vector-set! t 4 reconciled?)
      t )))

(define (transaction->amount t) (vector-ref t 0))
(define (transaction->date t) (vector-ref t 1))
(define (transaction->description t) (vector-ref t 2))
(define (transaction->other-account t) (vector-ref t 3))
(define (transaction->reconciled? t) (vector-ref t 4))

(define (sort-by-date l)
  (sort:sort l (lambda (a b) (date-less? (transaction->date a)
					 (transaction->date b)))))
	
; unique check mark token
(define ^ '^)

(define (make-account name description)
  (let ((transactions '()))

    (define (add! reconciled?)
      (lambda (howmuch when why other)
	(let ((t (make-transaction howmuch when why other reconciled?)))
	  (set! transactions (cons t transactions)))))

    (define (self reconciled?)
      (lambda (msg)
	(case msg
	  ((name) name)
	  ((description) description)
	  ((transactions) transactions)
	  ((^) (self #t))
	  ((add!) (add! reconciled?)))))
    
    (self #f)))

(define (make-combined-account name description . accounts)
  (define (acc->name acc) (acc 'name))
  (let ((acc-names (map acc->name accounts)))
    (define (transactions)
      (filter (lambda (t) 
		(not (memq (acc->name (transaction->other-account t)) 
			   acc-names)))
	      (apply append (map (lambda (acc) (acc 'transactions))
				 accounts))))
    (define self
      (lambda (msg)
	(case msg
	  ((name) name)
	  ((description) description)
	  ((transactions) (transactions)))))

    self))

(define (account->trasactions preds account)
  (filter* preds (account 'transactions)))

;; list of accounts
(define *accounts* '())
(define (register-account! a)
  (set! *accounts* (cons a *accounts*)))
(define (destroy-accounts!)
  (set! *accounts* '()))

(define-syntax define-account 
  (syntax-rules ()
    ((define-account <name> <description>)
     (begin
       (define <name> (make-account (quote <name>) <description>))
       (register-account! <name>)))))

(define-syntax define-combined-account
  (syntax-rules ()
    ((define-combined-account <name> <description> <accounts> ...)
     (define <name> (make-combined-account (quote <name>)
					   <description>
					   <accounts> ...)))))
       
;; user interface:
(define (move! when howmuch why from to)
  (howmuch 
   (lambda (in out)
     (begin ((from 'add!) (list out) when why to)
	    ((to 'add!) (list in) when why from)))))

(define (move*! when howmuch why accounts)
  (if (and (pair? accounts) (pair? (cdr accounts)))
      (begin
	(move! when howmuch why (car accounts) (cadr accounts))
	(move*! when howmuch why (cdr accounts)))))

(define (month mm)
  (lambda (dd yyyy)
    (let ((when (make-date yyyy mm dd)))
      (lambda (howmuch why acc0 acc1 . accn)
	(move*! when howmuch why (cons acc0 (cons acc1 accn)))))))

; user-visible date constructors.  Usage: (apr 1 2001)
(define jan (month 0))
(define feb (month 1))
(define mar (month 2))
(define apr (month 3))
(define may (month 4))
(define jun (month 5))
(define jul (month 6))
(define aug (month 7))
(define sep (month 8))
(define oct (month 9))
(define nov (month 10))
(define dec (month 11))

; user-visible value constructors.  Usage: (USD 10)
(define USD (make-currency 'USD))
(define Eur (make-currency 'Eur))

(define (money->string v)
  (let ((currency (car v))
	(amount (cdr v)))
    (case (car currency)
      ((in)  (sprintf #f "in  %-4a %9.2f" (cdr currency) amount))
      ((out) (sprintf #f "out %-4a %9.2f" (cdr currency) amount)))))

(define (money->algebraic v)
  (let ((currency (car v))
	(amount (cdr v)))
    (case (car currency)
      ((in)  (cons (cdr currency) amount))
      ((out)  (cons (cdr currency) (- amount))))))
      

(define report-transaction
  (lambda (transaction)
    (let ((date (transaction->date transaction))
	  (amount (transaction->amount transaction))
	  (desc (transaction->description transaction))
	  (reconciled? (transaction->reconciled? transaction))
	  (other (transaction->other-account transaction)))
      (printf "%1s %04d/%02d/%02d %a %-18a %s\n" 
	      (if reconciled? "^" " ")
	      (date->yyyy date)
	      (+ 1 (date->mm date)) ; month human format
	      (date->dd date)
	      (apply string-append (map money->string amount))
	      (other 'name)
	      desc))))
  
(define (report account . interesting?)
  (for-each report-transaction
	    (sort-by-date (account->trasactions interesting? account))))


(define (balance account . interesting?)
  (map money->string
       (value/sum (map transaction->amount 
		       (account->trasactions interesting? account)))))

(define (algebraic-balance account . interesting?)
  (value/sum 
   (map (lambda (v) (map money->algebraic v))
	(map transaction->amount 
	     (account->trasactions interesting? account)))))

(define (compose f g) (lambda (x) (f (g x))))

;; commonly used predicates
(define reconciled? transaction->reconciled?)
(define not-reconciled? (compose not reconciled?))

(define (before yyyy mm dd)
  (lambda (t)
    (date-less? (transaction->date t) (make-date yyyy (- mm 1) dd))))